cornish_fastb = """
@article{Cornish:2007if,
    author = "Cornish, Neil J. and Littenberg, Tyson B.",
    title = "{Tests of Bayesian Model Selection Techniques for Gravitational Wave Astronomy}",
    eprint = "0704.1808",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.76.083006",
    journal = "Phys. Rev. D",
    volume = "76",
    pages = "083006",
    year = "2007"
}
"""

robson_triple = """
@article{Robson:2018svj,
    author = "Robson, Travis and Cornish, Neil J. and Tamanini, Nicola and Toonen, Silvia",
    title = "{Detecting hierarchical stellar systems with LISA}",
    eprint = "1806.00500",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.98.064012",
    journal = "Phys. Rev. D",
    volume = "98",
    number = "6",
    pages = "064012",
    year = "2018"
}
"""

zenodo = """
@software{michael_l_katz_2022_6500434,
  author       = {Michael L. Katz},
  title        = {mikekatz04/GPUBACKENDTOOLS: First official public release!},
  month        = apr,
  year         = 2022,
  publisher    = {Zenodo},
  version      = {v1.0.0},
  doi          = {10.5281/zenodo.6500434},
  url          = {https://doi.org/10.5281/zenodo.6500434}
}
"""
